#Rem

Mini PWM4.1 PV charge controller
05 Oct 2023
Designer: Mike Scaife, NZ
for small PV panels 10-50 watts charging a low AHr 12\24 volt battery, either Lithium or Lead.


PV set pot adjusted for 30 volts PV input = ADC reading 1000
Bat volts pot set for 15 volts input = ADC reading 1000

PV panel used for testing was 30W  22v unloaded, 18v mppt
Battery type used in test gell 12v 8 AH battery
Bulk voltage = 14.6 volts
Float voltage = 13.6 volts

CPU IO
c.0=out {serialOut} with link for charge status Led
c.1=out(charge mosfets)
c.2= Adc Input(adc Battery Sense)
c.3=input with pullup (load switch mosfet, On = pullup present.,Off = no pullup)
c.4=Adc input(PVsense)
c.5=input {serialIn}

08M2, design has minimal IO capability, so no current sensing, to measure PV voltage the charge mosfets must be briefly turned off to unload the PV
if the PV voltage is higher than the battery then there will be some charge current into the battery.

#EndRem


#picaxe 08M2
#No_data
#Terminal  19200

'for testing, send text to the terminal window, turn off for Live, set to 0, this activates the status led when link on pcb made.
'0 = serial off except for calibratio9n mode
#define SerialTerminalOut 0


Setfreq m16

symbol ccBatBulkVolts = 973			'14.6 (14.6/15 ) * 1000
symbol ccBatFloatVolts = 906			'13.6	(13.6/15) * 1000
symbol ccConstantVoltageTime = 360		'720 x 10 sec, = 2 hours, for lifepo4 this will be less, give enough time for bms to top balance.
symbol ccPVLowTimeDelay = 60			'60 x 10 sec, = 10 minutes, effectively end of days charging.
symbol ccMinBatVoltsTimeDelay = 6		'6 x 10 sec = 60 sec, ok for lead acid, for Lithium make this lower eg 10 sec
symbol ccReturnBatVoltsTimeDelay = 3	'30 sec, time for loads to switch back on after min level battery volts

symbol ccBatMinVolts = 766			'11.5v if battery voltage drops to this level then turn off load mosfet
symbol ccBatMaxVolts = 1014			'15.2 if battery voltage increases to this point, Fault or incorrect initial calibration

symbol ccMosfetSwitchDelay = 20		'@ CPU clock 16 MHz, 20 = 5 miliseconds, allow time for slow voltage output  mosfet drivers to switch on\off

symbol ccState_Waiting = 0
symbol ccState_Bulk = 1
symbol ccState_ConstVoltage = 2
symbol ccState_Float = 4
symbol ccState_Fault_HighVolts = 8
symbol ccState_Fault_LowVolts = 16
symbol ccState_Calibrate = 32


symbol statusLed = c.0		'serial out pin, manual jumper to connect the led
symbol chargeOutMosfet = c.1
symbol batterySenseADC = c.2
symbol loadOnMosfet = c.3
symbol pvSenseADC = c.4
symbol calibrateSwitch = pinc.5	'serial input input, when disconnected used as calibration mode input

symbol calibrateOn_flag = bit24
symbol rulesLoadOn_flag = bit25
symbol statusFlash_flag = bit26
symbol loadOn_flag = bit27
symbol chargeOn_flag = bit28
symbol lowBattery_flag = bit29
symbol pvOK_flag = bit30
symbol highBattery_flag = bit31


symbol btemp1 = b4
symbol b2SecDelay = b5
symbol bChargeState = b6
symbol bRulesCounter = b7
symbol bminimumBatVoltsTimer = b8
symbol breturnBatVoltsTimer = b9
symbol wPVLowTimer = w9				'pv voltage less than battery voltage time out
symbol wConstantVoltageTimer = w10
symbol wtemp = w11
symbol wpvVolts = w12
symbol wbatVolts = w13


'turn On the status Led
#Macro TurnOnStatusLed
	High StatusLed
#EndMacro

'turn off the status Led
#Macro TurnOffStatusLed
	Low StatusLed
#EndMacro

'pulse the status led briefly
#Macro FlashStatusLed
	pulsout statusled, 5000
#EndMacro

'Load mosfet driven by internal pullup resistor on input pin c.3
'turn pullups off to disable loads
#Macro TurnOffLoad
	pullup OFF
	loadOn_Flag = 0
#EndMacro

'Turn on by enabling c.3 internal pullup
#Macro TurnOnLoad
	pullup %01000
	loadOn_Flag = 1
#EndMacro

'dual back to back mosfets, driven by low current isolated voltage driver
'slow on time approx 4 mSec, faster off time of 2mSec
'at each transition allow a delay of 5 mSec prior to measuring bat volts etc.
#Macro TurnOnChargeMosfet
	high chargeOutMosfet
	pause ccMosfetSwitchDelay
	chargeOn_flag = 1
#EndMacro

#Macro TurnOffChargeMosfet
	low chargeOutMosfet
	pause ccMosfetSwitchDelay
	chargeOn_flag = 0
#EndMacro






StartMain:
dirsC = %00000011
outpinsC = %00000000
adcsetup = %10100 ; set ADC2,4 for analog, note this disconnects the digital input connection to these pins

;set initial state
bChargeState = ccState_Waiting
time = 0
highBattery_flag = 0
lowBattery_flag = 0
calibrateOn_flag = 0
pvOK_flag = 0
wPVLowTimer = 0
bminimumBatVoltsTimer = 0
breturnBatVoltsTimer = 0
TurnOffLoad
TurnOffStatusLed
TurnOffChargeMosfet
'allow use of C.5 serial input pin as an input
disconnect

'Main Battery Charge State loop
do

'State code for battery charging progress
	Select Case bChargeState
		Case ccState_Waiting
		'waiting for PV voltage to > battery 
			gosub Waiting
			
		Case ccState_Bulk
		'Bulk charge voltage, panel is connected to the battery continuously
		'until voltage rises, then switched to Constant Voltage state
			gosub BulkCharge
			
		Case ccState_ConstVoltage
		'if bat volts exceed the Bulk set point the charge mosfets are turned off
		'turned back on when volts fall, ie pulsed current, pulse speed depends on the battery load.
		'held in this state for fixed duration, Gell Lead Bat 2 hrs, Lifepo4 10-20min to allow time to top balance
			gosub CVCharge
			
		Case ccState_Float
		'state following CV, remains in this until PV volts drop below battery
			gosub FloatCharge
			
	Case ccState_Fault_HighVolts
		'will only occur if the battery voltage goes above the max setting
		gosub FaultHighVolts
		'exit do loop and reset
		exit
	
	Case ccState_Fault_LowVolts
	'low battery volts, 
		gosub FaultLowVolts
		'exit do loop and reset
		exit
		
	Case ccState_Calibrate
	'user has switched serial input pin to +5v and entered calibrate mode
		gosub CalibrationMode
		
	end Select
	

loop
'if fault occurs restart
Reset


CalibrationMode:
'Note to exit this mode, turn the boards power off for a few seconds and re-apply
'turn off everything
TurnOffStatusLed
TurnOffLoad
TurnOffChargeMosfet

'read the Battery and PV voltages
'aim is to show a 1000 unit adc value for a set input
'12v battery, feed 15v psu and set pot for 1000
'21v PV, use 30v PSU, set pot for reading 1000

'Other battery 24v and 36V PV adjust accordinly

'normal mode serial is off = 0, but leave this active
#if SerialTerminalOut = 0
	sertxd("Calibrate:",cr,lf)
	do
		gosub Readbattery
		gosub ReadPV
		pause 400
		sertxd("Battery Read: ",#wBatVolts,cr,lf,"PV Read: ",#wpvVolts,cr,lf,cr,lf)

	loop
#endif


return




Waiting:
'wait for PV voltage to exceed battery and any high voltage flags to clear
'turn off charge mosfets
	#if SerialTerminalOut = 0
		TurnOffStatusLed
	#endif
	TurnOffChargeMosfet
	#if SerialTerminalOut = 1
		sertxd("Waiting: PV_Flag: ",#pvOK_flag,cr,lf)
	#endif
	
	#if SerialTerminalOut = 0
		'use pulse out to briefly flash the led once every 3 sec
		b2SecDelay = time//3
		if b2SecDelay = 0 then
			if statusFlash_flag = 0 then
				FlashStatusLed
				statusFlash_flag = 1
			end if
			
		else
			statusFlash_flag = 0
			
		end if
	#endif

	'check battery and PV, change state to Bulk if OK
	gosub ReadBattery
	gosub CheckRules
	
	'if calibration mode detected exit waiting
	if bChargeState = ccState_Calibrate then
		return
		
	elseif pvOK_flag = 1 AND highBattery_flag = 0 then
		'wake up and start charging
		bChargeState = ccState_Bulk
	end if
	

'returns to main state loop, runs the loop and repeats until Bulk begins
return



BulkCharge:
;'Bulk charge mode, essentially no PWM, mosfets turned on continuously until set voltage reached or fault
;'note keep battery cables sized according to current to avoid voltage drops causing incorrect voltage readings
;'eg 5 amps use > 4mm^2 wire
#if SerialTerminalOut = 1
	sertxd("Starting Bulk: ",cr,lf)
#endif
time = 0
bRulesCounter = 0
#if SerialTerminalOut = 0
	TurnOffStatusLed
#endif
TurnOnChargeMosfet
gosub ReadBattery

do while wbatVolts < ccBatBulkVolts AND bChargeState = ccState_Bulk
	
	gosub ReadBattery	'battery over volts also tested
	'every 10 seconds, rules turn off charge mosfets and measure PV etc
	'any errors the State will be changed to something else
	gosub CheckRules
	
	#if SerialTerminalOut = 1
		sertxd("Bulk Charging: ",#wbatVolts,cr,lf)
	#endif
	#if SerialTerminalOut = 0
		TurnOnStatusLed
	#endif
loop
'Bulk done or error fault, turn off charge mosfets

#if SerialTerminalOut = 0
	TurnOffStatusLed
#endif
TurnOffChargeMosfet

'various fault flags set
if highBattery_flag = 1 then
	#if SerialTerminalOut = 1
		sertxd("Bulk Fault, High Battery Volts: ",#wbatVolts,cr,lf)
	#endif

else if pvOK_flag = 0 then
	#if SerialTerminalOut = 1
		sertxd("Bulk Fault, Low PV: ",#wpvVolts,cr,lf)
	#endif

else if lowBattery_flag = 1 then
	#if SerialTerminalOut = 1
		sertxd("Bulk Fault, Low Battery: ",#wbatVolts,cr,lf)
	#endif
end if
	
if bchargeState = ccState_Bulk then
'normal, bulk complete, change to constant volts	
	bChargeState = ccState_ConstVoltage
	#if SerialTerminalOut = 1
		sertxd("Bulk Ending, move to CV: ",#wbatVolts,cr,lf)
	#endif
end if

'return with state flag set
Return
			


CheckRules:

'gets repeatedly called, after 10 seconds then if the charge mosfet is on, turn off to prevent the extra time delay increasing the battery voltage
'pvOK_flag set to 0 if low PV
'lowBattery_flag = 0 if battery volts are low
'loadOn_flag = 0 if external load are turned off
'wConstantVoltageTimer incremented every 10 sec



if time > 10 then
		rulesLoadOn_flag = chargeOn_flag
		'if the charge mosfets were on at time, then briefly switch off
		if rulesLoadOn_flag = 1 then
			TurnOffChargeMosfet
		end if
	
		'check the pv voltages, wpvVolts set by call
		'when in CV stage this timer called every 20 sec sets the CV Duration
		inc wConstantVoltageTimer
		gosub ReadPV
		
		'if the pv voltage > battery voltage + a margin, then set the PVOK_flag to 1
		'ie pv can charge the battery, at night it will be way less
		wtemp = wpvVolts - 20
		if wtemp > wbatVolts then
			pvOK_flag = 1
			wPVLowTimer = 0
		else
			'pv has to be low for a set period before tripping flag to 0
			'only want it to trip if sun has gone.
			if wPVLowTimer < ccPVLowTimeDelay then
				inc wPVLowTimer
				#if SerialTerminalOut = 1
					sertxd("Low PV Timing out: ",#wtemp, " Bat Volts: ",#wbatVolts,", Delay Counter: ",#wPVLowTimer,cr,lf)
				#endif
			else
			'delay expired, pv is missing so no charge possible, set status to waiting
				pvOK_flag = 0
				bchargeState = ccState_Waiting
				#if SerialTerminalOut = 1
					sertxd("No PV, State changed to Waiting:",cr,lf)
				#endif
			end if
				
		end if
	
	
		'test the load on battery voltage, if the battery is less than a value that would damage it
		'then turn the load off after a short delay, as a heavy short load may pull the volts lower
		'and cause false tripping.
		if wbatVolts <= ccBatMinVolts then
			if bminimumBatVoltsTimer < ccMinBatVoltsTimeDelay then
				inc bminimumBatVoltsTimer
				#if SerialTerminalOut = 1
					sertxd("Low Battery Detected, Timing: ",#wbatVolts, ", Delay: ",#bminimumBatVoltsTimer,cr,lf)
				#endif
				
				
			else
				'if delay has passed and battery volts still low then turn loads off, raise alarm
				TurnOffLoad
				#if SerialTerminalOut = 1
					sertxd("Load Turned OFF, Low Battery: ",#wbatVolts,", Set Point: ",#ccBatMinVolts, cr,lf)
				#endif
				
				lowBattery_flag = 1
				'todo Alarm, no IO pins for an alarm, need CPU with more pins
				'use low current relay across load, so when it drops out contact closes turns on piezo siren
				bchargeState = ccState_Fault_LowVolts
				
			end if
		
		
			'reset return timer
			breturnBatVoltsTimer = 0
			
		else
			'battery volts ok so turn on the pullup resistor on input c.3
			'this will switch the schmitt trigger inverter input high, output goes low
			'and turns on the opto driver for load mosfet
			'turn it on if voltage stable for set period
			if lowBattery_flag = 1 then
				if breturnBatVoltsTimer < ccReturnBatVoltsTimeDelay then
					inc breturnBatVoltsTimer
					#if SerialTerminalOut = 1
						sertxd("Low Battery Return Delay: ",#wbatVolts, ", Delay: ",#breturnBatVoltsTimer,cr,lf)
					#endif
				else 
					'turn load back on
					TurnOnLoad
					lowBattery_flag = 0
					#if SerialTerminalOut = 1
						sertxd("Load On after Low Battery: ", cr,lf)
					#endif
					
				end if
				
			
			elseif lowBattery_flag = 0 AND loadOn_flag = 0 then
				'turn load back on, 
				TurnOnLoad
				#if SerialTerminalOut = 1
					sertxd("Load On: ", cr,lf)
				#endif
				
			end if
			'reset voltage low timer
			bminimumBatVoltsTimer = 0
		end if
	
	'if the charge mosfet was on at the start, turn it back on
	if rulesLoadOn_flag = 1 then
		TurnOnChargeMosfet
	end if
	
	'reset time 
	time = 0
	
	'test the c.5 pin for a logic 1, if contact made > 10sec, switch to calibrate mode
	'this locks In until reset
	if calibrateSwitch = 1 or calibrateOn_flag = 1 then
		bchargeState = ccState_Calibrate
		calibrateOn_flag = 1
	end if
	
end if


Return



CVCharge:
'setup a time delay to remain in CV mode for fixed duration
'if voltage = or exceeds Bulk value then turn charge off, until voltage drops 
'repeat

time = 0
wConstantVoltageTimer = 0
#if SerialTerminalOut = 1
	sertxd("CV Started: ", #wbatVolts, ", Set Val: ", #ccBatBulkVolts,cr,lf)
#endif

'exit on time out or state has changed
do while wConstantVoltageTimer < ccConstantVoltageTime AND bChargeState = ccState_ConstVoltage
	
	for bRulesCounter = 0 to 255
		gosub ReadBattery
		if wbatVolts < ccBatBulkVolts then
			TurnOnChargeMosfet
			#if SerialTerminalOut = 0
				select case bRulesCounter
					case < 30
						TurnOnStatusLed
					else
						TurnOffStatusLed
				end select
			#endif
			#if SerialTerminalOut = 1
				sertxd("CV ON: ", #wbatVolts,cr,lf)
			#endif
		else
			TurnOffChargeMosfet
			#if SerialTerminalOut = 0
				TurnOffStatusLed
			#endif
			#if SerialTerminalOut = 1
				sertxd("CV OFF: ", #wbatVolts,cr,lf)
			#endif
	
		end if
	
	next
	TurnOffChargeMosfet
	gosub ReadBattery
	gosub CheckRules
	
loop	
	

TurnOffChargeMosfet
#if SerialTerminalOut = 0
	TurnOffStatusLed
#endif

'various fault flags set
if highBattery_flag = 1 then
	#if SerialTerminalOut = 1
		sertxd("CV Fault, High Battery Volts: ",#wbatVolts,cr,lf)
	#endif

else if pvOK_flag = 0 then
	#if SerialTerminalOut = 1
		sertxd("CV Fault, Low PV: ",#wpvVolts,cr,lf)
	#endif

else if lowBattery_flag = 1 then
	#if SerialTerminalOut = 1
		sertxd("CV Fault, Low Battery: ",#wbatVolts,cr,lf)
	#endif
end if
	
if bchargeState = ccState_ConstVoltage then
'normal, CV complete, change to Float volts	
	bChargeState = ccState_Float
	#if SerialTerminalOut = 1
		sertxd("CV Ending, move to Float: ",#wbatVolts,cr,lf)
	#endif
end if


Return


FloatCharge:
'remain in float mode untill loss of PV; usually at night.
'Note: you could add code here to detect a drop in battery voltage then switch directly back to Bulk
'found not really necessary.
'When the volts drop, mosfets turn on for longer or stay on continuously, slow pwm cycle
'and charge at what ever current the PV will supply.

'charge mosfets are switched on-off as battery voltage rises\falls with pv current and battery load
'not fixed pwm frequency, as switching freq determind by speed of code execution and pv current.
'The isolated mosfet drivers convert drive current into an output voltage, approx 10v to drive mosfet gate
'these are quite slow ie 5ms to turn on 2ms to turn off, so a fast cpu pwm output port is not used.


#if SerialTerminalOut = 1
	sertxd("Float Started: ", #wbatVolts, ", Set Val: ", #ccBatFloatVolts,cr,lf)
#endif
'reset the rules start time to 0
time = 0

'indefinite loop, will only exit if charge state <> float, ie no pv, bat volts too low\high
'on dusk as sun goes down, the no pv rules will change state and exit float back to waiting.
'exit on time out or state has changed
do while bChargeState = ccState_Float
	
	'every 256 loops call the rules
	for bRulesCounter = 0 to 255
		gosub ReadBattery
		if wbatVolts < ccBatFloatVolts then
			TurnOnChargeMosfet
			#if SerialTerminalOut = 0
				FlashStatusLed
			#endif
		
			#if SerialTerminalOut = 1
				sertxd("Float ON: ", #wbatVolts,cr,lf)
			#endif
		else
			TurnOffChargeMosfet
			
			#if SerialTerminalOut = 1
				sertxd("Float OFF: ", #wbatVolts,cr,lf)
			#endif
	
		end if
	
	next
	TurnOffChargeMosfet
	gosub ReadBattery
	gosub CheckRules
	
loop	


'exit with mosfets off
TurnOffChargeMosfet

#if SerialTerminalOut = 0
	TurnOffStatusLed
#endif

'various fault flags maybe set
if highBattery_flag = 1 then
	#if SerialTerminalOut = 1
		sertxd("Float Fault, High Battery Volts: ",#wbatVolts,cr,lf)
	#endif

else if pvOK_flag = 0 then
	#if SerialTerminalOut = 1
		sertxd("Float Fault, Low PV: ",#wpvVolts,cr,lf)
	#endif

else if lowBattery_flag = 1 then
	#if SerialTerminalOut = 1
		sertxd("Float Fault, Low Battery: ",#wbatVolts,cr,lf)
	#endif
end if
	




Return


FaultHighVolts:
'High battery voltage fault
'
'make sure to turn off charge mosfets
TurnOffChargeMosfet

'turn off load, as it could be damaged by high DC volts
TurnOffLoad

#if SerialTerminalOut = 1
	sertxd("Fault Battery Volts High: ", #wbatVolts, ", Set Val: ", #ccBatMaxVolts,cr,lf)
#endif
	

Return

FaultLowVolts:
'Low battery voltage fault
'make sure any loads are removed
'turn off loads
TurnOffLoad
sleep 10

Return


ReadPV:
'prior calls have turned off charge to unload PV voltage as it is clamped to the battery voltage
'take 4 readings of PV voltage and average
wpvVolts = 0
for btemp1 = 0 to 3
	readadc10 pvSenseADC,wtemp
	wpvVolts = wpvVolts + wtemp

next
'take average
wpvVolts = wpvVolts / 4

'calibrated for 30v PV in reads adc = 1000, actual pv = approx 22v unloaded, 18v mppt
'so actual PV =  (ADC reading/1000) * 30
'if other PV voltages used then change calibration 
wtemp = wpvVolts

'now scale PV * 2 to directly map to battery volts with ADC of 1000 = 15v
'then we can direcly compare battery against PV
'dont go below 30 so later code wont go below 0
wpvVolts = wpvVolts * 2 Min 30

#if SerialTerminalOut = 1
	sertxd("Read PV, Actual: ",#wtemp, ", Scaled: ",#wpvVolts, cr,lf)
#endif

Return



ReadBattery:
'take 4 readings of battery voltage and average
'also set battery High flag
'wbatVolts set to value

wbatVolts = 0
for btemp1 = 0 to 3
	readadc10 batterySenseADC, wtemp
	wbatVolts = wbatVolts + wtemp
next
'take average
wbatVolts = wbatVolts / 4

'check for battery too high fault
'set flag and change state, this is read continuously
'so fast reaction to change
if wbatVolts >= ccBatMaxVolts then
'set over voltages flag
	highBattery_flag = 1
	bChargeState = ccState_Fault_HighVolts
else
	highBattery_flag = 0
end if

return
